# Written by Eckhard Pfluegel, 
polysols is taken from ratsols by Paul Zimmermann #

# ode::polysols(lode, x, N)                                                   
    INPUT:  lode - list of coefficients of linear differential operator
		   (or linear ODE from toplevel)
            x    - a name (or y(x) from toplevel))
            N    - (optional) an integer                                      
    OUTPUT: a basis of the space of all polynomial solutions of lode. 
            If N is passed as argument, it will be used as upper bound for 
            the degree.                        
Example:
>> ode::polysols(-8*y(x)+8*x*diff(y(x),x)+156*x^2*diff(y(x),x$2)
	+258*x^3*diff(y(x),x$3)+122*x^4*diff(y(x),x$4)
	+20*x^5*diff(y(x),x$5)+x^6*diff(y(x),x$6),y(x));

                                    {x}

>> ode::polysols([-8,8*x,156*x^2,258*x^3,122*x^4,20*x^5,x^6],x);

                                    {x}

>> ode::polysols([12*x^3+6*x^2-2*x-3,8*x^4+2*x^3-5*x^2-x,x^5-x^3],x);

                                    {}
#                                                                             
ode::polysols := proc(lode, x, N)
    local np, u, q, N1, R, i, a, Q, n, sys, unk, j, e, eq, res, v;
begin
    if testargs() then
       if type(lode)=DOM_LIST then #ok#
       elif (lode:=ode::isLODE(lode,x,"Lode"))=FALSE then
          error("not a linear ODE with rational coefficients")
       else
          x:=lode[3];
          lode:=[coeff(lode[1],[diff(args(2),x$i)],1)$i=0..lode[4]];
       end_if
    end_if;
    np := ode::newtonPolygon(lode, x, 1/x, u);
    n := nops(lode);
    q:=sharelib::iroots(np[3][1][2], FALSE);
    if q={} then q:=0 else q:=min(op(q)) end_if;
    if args(0) = 3 then N1 := N else N1 := max(n-1,-q) end_if;
    R:=0; for i from 0 to N1 do a[i]:=genident("a"); R:=R+a[i]*x^i end_for;
    Q[1]:=R; for i from 2 to n do 
        Q[i]:=normal(diff(Q[i-1],x))
    end_for;
    sys:=numer(_plus(expr(lode[i])*Q[i]$hold(i)=1..n));
    sys:={coeff(sys,[x])}; unk:={a[i]$hold(i)=0..N1};
    # the following lines avoid an expensive call to linsolve #
    j:=null();
    repeat
        e:=FALSE; 
        for eq in sys do
            i:=indets(eq) intersect unk;
            if nops(i)=1 then
                i:=op(i);
                if degree(eq,[i])=1 then
                    sys:=sys minus {eq}; unk:=unk minus {i};
                    eq:=i=-coeff(eq,[i],0)/coeff(eq,[i],1);
                    j:=j,eq; sys:=subs(sys,eq); e:=TRUE; break;
                end_if
            end_if
        end_for;
   until e=FALSE end_repeat;
   sys:=sys minus {0};
   if sys<>{} then
       sys:=linsolve(sys,unk);
       if sys=null() or sys=FAIL then return({}) end_if;
       j:=j,op(sys)
   end_if;
   R := subs(R,j);
   # I joined this for the moment #
   if R = {} then return({}) end_if;
   e := indets(R) minus {x};
   if nops(e) = 1 then return({subs(R, op(e) = 1)}) end_if; 
   res := {};
   for i in e do
       res := res union {subs(R, {seq(v=0, v=(e minus {i})), i=1})}
   end_for;
   res;
end_proc:

# Tools for computing Newton polygons and valuations #

#
 ode::newtonPolygon(lode, x, p, u[, m])                                      
    INPUT:  lode - list of coefficients of linear differential operator       
            x    - a name                                                     
            p    - a polynomial or 1/x                                        
            u    - a name                                                     
            m    - (optional) an integer                                      
    OUTPUT: the Newton polygon and of lode at place p as list                 
            [minimal y-value, u, list of couples [slope, newton polynomial]]  
            the polygon is constructed only for lode[i], i = 0,..,m           
#                                                                             
ode::newtonPolygon := proc(lode, x, p, u, m)
    local lode1, n, s, yp, lcs, drop, npoly, d, i, j, k, slope, tmp, res, dp;
begin
    lode1 := lode;
    if args(0) = 5 then n := m+1 else n := nops(lode) end_if;
    if p = 1/x then s := -1; dp := 1 else s := 1; dp := diff(p, x) end_if;
    yp := []; lcs := [];
    for i from 1 to n do
       tmp := ode::valuation(lode1[i], x, p);
       yp  := append(yp, tmp[1]-s*(i-1));
       lcs := append(lcs, tmp[2])
    end_for;
    drop := min(op(yp));
    npoly := 0;
    for j from 1 to n do
        if yp[j] = drop then  
            npoly := npoly + ode::tau(lcs[j], x, p, j-1)*
                             _mult(s*u-k+1$hold(k)=1..j-1)
        end_if
    end_for;
    d := degree(npoly, [u]);
    if d > 0 then res := [[0, collect(normal(primpart(npoly, [u])), [u])]] 
        else res := [] 
    end_if;
    i := d+1;
    while i < n do
        slope := min(seq((yp[j]-yp[i])/(j-i), j=[$(i+1..n)]));
        npoly := 0;
        for j from i to n do
            if yp[j] -(j-i)*slope = yp[i] then  
                npoly := npoly + ode::tau(lcs[j], x, p, j-i)*
                                 (s*u)^((j-i)/denom(slope))
            end_if
        end_for;
        res := append(res, [slope,collect(normal(primpart(npoly, [u])), [u])]);
        i := i+degree(npoly, [u])*denom(slope);
    end_while;
    [drop, u, res] 
end_proc:

# ode::valuation(a, x, p)                                                     
    INPUT : a - a polynomial
            x - a name
            p - a polynomial or 1/x                
    OUTPUT: the list [v, lc] where v is the p-valuation of f and lc the       
            leading coefficient of a at p.                                    
#
ode::valuation:=proc(a, x, p)
    local ord, tmp;
begin
    if (a = 0) then 
        return([infinity, 0])
    elif (p = 1/x) then
        return([-degree(a, x), lcoeff(a, [x])])
    elif (p = x) then
        return([ldegree(a, x), tcoeff(a, [x])])
    end_if;
    ord := 0; tmp := a;
    repeat
        tmp := divide(a, p, [x]);
        if iszero(tmp[2]) then ord := ord+1; a := tmp[1] else break end_if;
    until FALSE end_repeat;
    [ord, divide(a, p, [x], Rem)]
end_proc:

# ode::tau(lc, x, p, i)                                                       
    INPUT:  lc - a univariate polynomial in x of degree < degree(p)           
            x  - a name                                                       
            p  - an irreducible polynomial in x                               
            i  - an integer                                                   
    OUTPUT: the ith remainder of a polynomial whose leading coefficient       
            is lc at p                                                        
#
ode::tau := proc(lc, x, p, i)
begin   
    if (p = x) or (p = 1/x) then return(lc)
        else return(divide(lc*diff(p, x)^i, p, [x], Rem))
    end_if
end_proc:

